"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Slice = void 0;
const cell_1 = require("./cell");
const coins_1 = require("../coins");
const address_1 = require("../address");
const helpers_1 = require("../utils/helpers");
class Slice {
    constructor(bits, refs) {
        this._bits = bits;
        this._refs = refs;
    }
    get bits() {
        return Array.from(this._bits);
    }
    get refs() {
        return Array.from(this._refs);
    }
    static bitsToBigUint(bits) {
        if (!bits.length)
            return { value: 0n, isSafe: true };
        const value = bits
            .reverse()
            .reduce((acc, bit, i) => (BigInt(bit) * (2n ** BigInt(i)) + acc), 0n);
        const isSafe = value <= Number.MAX_SAFE_INTEGER;
        return {
            value,
            isSafe
        };
    }
    static bitsToBigInt(bits) {
        if (!bits.length)
            return { value: 0n, isSafe: true };
        const { value: uint } = Slice.bitsToBigUint(bits);
        const size = BigInt(bits.length);
        const int = 1n << (size - 1n);
        const value = uint >= int ? (uint - (int * 2n)) : uint;
        const isSafe = value >= Number.MIN_SAFE_INTEGER && value <= Number.MAX_SAFE_INTEGER;
        return {
            value,
            isSafe
        };
    }
    static bitsToIntUint(bits, options) {
        const { type = 'uint' } = options;
        const { value, isSafe } = type === 'uint'
            ? Slice.bitsToBigUint(bits)
            : Slice.bitsToBigInt(bits);
        if (!isSafe) {
            throw new Error('Slice: loaded value does not fit max/min safe integer value, use alternative BigInt methods.');
        }
        return Number(value);
    }
    skip(size) {
        if (this._bits.length < size) {
            throw new Error('Slice: skip bits overflow.');
        }
        this._bits.splice(0, size);
        return this;
    }
    skipDict() {
        this.loadDict();
        return this;
    }
    loadRef() {
        if (!this._refs.length) {
            throw new Error('Slice: refs overflow.');
        }
        return this._refs.shift();
    }
    preloadRef() {
        if (!this._refs.length) {
            throw new Error('Slice: refs overflow.');
        }
        return this._refs[0];
    }
    loadBit() {
        if (!this._bits.length) {
            throw new Error('Slice: bits overflow.');
        }
        return this._bits.shift();
    }
    preloadBit() {
        if (!this._bits.length) {
            throw new Error('Slice: bits overflow.');
        }
        return this._bits[0];
    }
    loadBits(size) {
        if (size < 0 || this._bits.length < size) {
            throw new Error('Slice: bits overflow.');
        }
        return this._bits.splice(0, size);
    }
    preloadBits(size) {
        if (size < 0 || this._bits.length < size) {
            throw new Error('Slice: bits overflow.');
        }
        return this._bits.slice(0, size);
    }
    loadInt(size) {
        const bits = this.loadBits(size);
        return Slice.bitsToIntUint(bits, { type: 'int' });
    }
    preloadInt(size) {
        const bits = this.preloadBits(size);
        return Slice.bitsToIntUint(bits, { type: 'int' });
    }
    loadBigInt(size) {
        const bits = this.loadBits(size);
        const { value } = Slice.bitsToBigInt(bits);
        return value;
    }
    preloadBigInt(size) {
        const bits = this.preloadBits(size);
        const { value } = Slice.bitsToBigInt(bits);
        return value;
    }
    loadUint(size) {
        const bits = this.loadBits(size);
        return Slice.bitsToIntUint(bits, { type: 'uint' });
    }
    preloadUint(size) {
        const bits = this.preloadBits(size);
        return Slice.bitsToIntUint(bits, { type: 'uint' });
    }
    loadBigUint(size) {
        const bits = this.loadBits(size);
        const { value } = Slice.bitsToBigUint(bits);
        return value;
    }
    preloadBigUint(size) {
        const bits = this.preloadBits(size);
        const { value } = Slice.bitsToBigUint(bits);
        return value;
    }
    loadVarInt(length) {
        const size = Math.ceil(Math.log2(length));
        const sizeBytes = this.loadUint(size);
        const sizeBits = sizeBytes * 8;
        return this.loadInt(sizeBits);
    }
    preloadVarInt(length) {
        const size = Math.ceil(Math.log2(length));
        const sizeBytes = this.preloadUint(size);
        const sizeBits = sizeBytes * 8;
        const bits = this.preloadBits(size + sizeBits).slice(size);
        return Slice.bitsToIntUint(bits, { type: 'int' });
    }
    loadVarBigInt(length) {
        const size = Math.ceil(Math.log2(length));
        const sizeBytes = this.loadUint(size);
        const sizeBits = sizeBytes * 8;
        const bits = this.loadBits(sizeBits);
        const { value } = Slice.bitsToBigInt(bits);
        return value;
    }
    preloadVarBigInt(length) {
        const size = Math.ceil(Math.log2(length));
        const sizeBytes = this.preloadUint(size);
        const sizeBits = sizeBytes * 8;
        const bits = this.preloadBits(size + sizeBits).slice(size);
        const { value } = Slice.bitsToBigInt(bits);
        return value;
    }
    loadVarUint(length) {
        const size = Math.ceil(Math.log2(length));
        const sizeBytes = this.loadUint(size);
        const sizeBits = sizeBytes * 8;
        return this.loadUint(sizeBits);
    }
    preloadVarUint(length) {
        const size = Math.ceil(Math.log2(length));
        const sizeBytes = this.preloadUint(size);
        const sizeBits = sizeBytes * 8;
        const bits = this.preloadBits(size + sizeBits).slice(size);
        return Slice.bitsToIntUint(bits, { type: 'uint' });
    }
    loadVarBigUint(length) {
        const size = Math.ceil(Math.log2(length));
        const sizeBytes = this.loadUint(size);
        const sizeBits = sizeBytes * 8;
        const bits = this.loadBits(sizeBits);
        const { value } = Slice.bitsToBigUint(bits);
        return value;
    }
    preloadVarBigUint(length) {
        const size = Math.ceil(Math.log2(length));
        const sizeBytes = this.preloadUint(size);
        const sizeBits = sizeBytes * 8;
        const bits = this.preloadBits(size + sizeBits).slice(size);
        const { value } = Slice.bitsToBigUint(bits);
        return value;
    }
    loadBytes(size) {
        const bits = this.loadBits(size);
        return (0, helpers_1.bitsToBytes)(bits);
    }
    preloadBytes(size) {
        const bits = this.preloadBits(size);
        return (0, helpers_1.bitsToBytes)(bits);
    }
    loadString(size = null) {
        const bytes = size === null
            ? this.loadBytes(this._bits.length)
            : this.loadBytes(size);
        return (0, helpers_1.bytesToString)(bytes);
    }
    preloadString(size = null) {
        const bytes = size === null
            ? this.preloadBytes(this._bits.length)
            : this.preloadBytes(size);
        return (0, helpers_1.bytesToString)(bytes);
    }
    loadAddress() {
        const FLAG_ADDRESS_NO = [0, 0];
        const FLAG_ADDRESS = [1, 0];
        const flag = this.preloadBits(2);
        if (flag.every((bit, i) => bit === FLAG_ADDRESS_NO[i])) {
            return this.skip(2) && address_1.Address.NONE;
        }
        if (flag.every((bit, i) => bit === FLAG_ADDRESS[i])) {
            const size = 2 + 1 + 8 + 256;
            const bits = this.preloadBits(size);
            const anycast = bits.splice(2, 1);
            const workchain = (0, helpers_1.bitsToInt8)(bits.splice(2, 8));
            const hash = (0, helpers_1.bitsToHex)(bits.splice(2, 256));
            const raw = `${workchain}:${hash}`;
            return this.skip(size) && new address_1.Address(raw);
        }
        throw new Error('Slice: bad address flag bits.');
    }
    preloadAddress() {
        const FLAG_ADDRESS_NO = [0, 0];
        const FLAG_ADDRESS = [1, 0];
        const flag = this.preloadBits(2);
        if (flag.every((bit, i) => bit === FLAG_ADDRESS_NO[i])) {
            return address_1.Address.NONE;
        }
        if (flag.every((bit, i) => bit === FLAG_ADDRESS[i])) {
            const size = 2 + 1 + 8 + 256;
            const bits = this.preloadBits(size);
            const anycast = bits.splice(2, 1);
            const workchain = (0, helpers_1.bitsToInt8)(bits.splice(2, 8));
            const hash = (0, helpers_1.bitsToHex)(bits.splice(2, 256));
            const raw = `${workchain}:${hash}`;
            return new address_1.Address(raw);
        }
        throw new Error('Slice: bad address flag bits.');
    }
    loadCoins() {
        const coins = this.loadVarUint(16);
        return new coins_1.Coins(coins, true);
    }
    preloadCoins() {
        const coins = this.preloadVarUint(16);
        return new coins_1.Coins(coins, true);
    }
    loadDict() {
        const isEmpty = this.preloadBit() === 0;
        if (isEmpty) {
            this.skip(1);
            return null;
        }
        return new cell_1.Cell([this.loadBit()], [this.loadRef()], false);
    }
    preloadDict() {
        const isEmpty = this.preloadBit() === 0;
        if (isEmpty) {
            return null;
        }
        return new cell_1.Cell([this.preloadBit()], [this.preloadRef()], false);
    }
    static parse(cell) {
        return new Slice(cell.bits, cell.refs);
    }
}
exports.Slice = Slice;
//# sourceMappingURL=slice.js.map