"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Mnemonic = void 0;
const tweetnacl_1 = __importDefault(require("tweetnacl"));
const crypto_js_1 = require("crypto-js");
const bip_0039_en_json_1 = __importDefault(require("./bip-0039-en.json"));
const helpers_1 = require("../utils/helpers");
const hash_1 = require("../utils/hash");
class Mnemonic {
    constructor(mnemonic = [], salt = null) {
        if (mnemonic.length && mnemonic.length !== 24) {
            throw new Error('Mnemonic: must contain 24 bip39 words.');
        }
        const words = mnemonic.length ? mnemonic : Mnemonic.genereteWords();
        const seed = Mnemonic.genereteSeed(words, salt);
        const sign = tweetnacl_1.default.sign.keyPair.fromSeed(seed);
        const keys = {
            private: sign.secretKey,
            public: sign.publicKey
        };
        this._words = words;
        this._seed = seed;
        this._keys = keys;
    }
    get words() {
        return this._words;
    }
    get seed() {
        return this._seed;
    }
    get keys() {
        return this._keys;
    }
    static deriveChecksumBits(entropy) {
        const CS = (entropy.length * 8) / 32;
        const hex = (0, hash_1.sha256)(entropy);
        const bits = (0, helpers_1.hexToBits)(hex);
        return bits.slice(0, CS);
    }
    static normalize(str) {
        return (str || '').normalize('NFKD');
    }
    static genereteWords() {
        const entropy = tweetnacl_1.default.randomBytes(32);
        const checkSumBits = Mnemonic.deriveChecksumBits(entropy);
        const entropyBits = (0, helpers_1.bytesToBits)(entropy);
        const fullBits = entropyBits.concat(checkSumBits);
        const chunks = fullBits.join('').match(/(.{1,11})/g);
        const words = chunks.map((chunk) => {
            const index = parseInt(chunk, 2);
            return bip_0039_en_json_1.default[index];
        });
        return words;
    }
    static genereteSeed(_mnemonic, _salt) {
        const PBKDF2_ROUNDS = 2048;
        _mnemonic.forEach((word) => {
            if (!bip_0039_en_json_1.default.includes(word)) {
                throw new Error(`Mnemonic: invalid mnemonic phrase words: "${word}".`);
            }
        });
        const mnemonic = Mnemonic.normalize(_mnemonic.join(' '));
        const salt = _salt !== null
            ? `mnemonic${Mnemonic.normalize(_salt)}`
            : 'mnemonic';
        const wordArray = {
            iterations: PBKDF2_ROUNDS,
            keySize: 512 / 32,
            hasher: crypto_js_1.algo.SHA512
        };
        const hex = (0, crypto_js_1.PBKDF2)(mnemonic, salt, wordArray).toString();
        const bytes = (0, helpers_1.hexToBytes)(hex);
        return bytes.slice(0, 32);
    }
}
exports.Mnemonic = Mnemonic;
//# sourceMappingURL=mnemonic.js.map